/* NLFILE.C - File functions/routines for Asgard-LU.C
Copyright 1988 by Brent Barrett
*/

#include "AsgardLU.h"
#include "extern.h"
#include "mem.h"

/* add_files()
add files to library
*/
add_files(mask)
    char *mask;
{
    char done;
    int i, j, p;
    unsigned int l, d;
    long fl;
    struct ffblk dir;

    if (!mask[0] || mask == 1) return;

    if (!library()) return;

    done = findfirst(mask, &dir, 0);

    while(!done) {

        if (stricmp(curr_file, dir.ff_name) != 0) {

            printf("  Adding: %12s (%ld bytes)", dir.ff_name, dir.ff_fsize);
            if (find_member(dir.ff_name, 0) == ERROR) {

                i = find_free();

                if (i == ERROR) {
                    printf("\n NO SPACE\n");
                    return;
                }

                clear_member(&entry[i], ACTIVE);
                fn_split(dir.ff_name, entry[i].name, entry[i].ext, FALSE);

                l = dir.ff_fsize/128;
                if (dir.ff_fsize > l * 128) {
                    l++;
                    entry[i].pad_count = ((l) * 128) - dir.ff_fsize;
                }

                if ((aux_fs=fopen(dir.ff_name, "rb"))==NULL) {
                    printf(" \n CAN NOT OPEN FILE!\n");
                    return;
                }

                fclose(LBR_fs);
                if ((LBR_fs=fopen(curr_file, "ab"))==NULL) {
                    printf(" CAN NOT APPEND TO LIBRARY!\n");
                    return;
                }

                entry[i].index = filelength(LBR_fs->fd)/128;
                entry[i].length = l;

                while(TRUE) {
                    strnset(sector_buf, 0x1A, 128);
                    l = fread(sector_buf, 128, 1, aux_fs);

                    if (fwrite(sector_buf, 128, 1, LBR_fs) != 1) {
                        printf(" \n ERROR WRITING! Disk full?\n");
                        fclose(aux_fs);
                        return;
                    }
                    if (l != 1) break;
                }
                fclose(aux_fs);
                putchar('\n');
                active_members++;

                fclose(LBR_fs);
                if ((LBR_fs=fopen(curr_file, "r+b"))==NULL) {
                    printf(" CAN NOT REOPEN LIBRARY!\n");
                    close_file();
                    return;
                }

                if (write_index(i) == ERROR) return;
            } else {
                printf(".. ALREADY IN LIBRARY, NOT ADDED\n");
            }
        }
        done = findnext(&dir);

    }

}

/* close_file()
closes current LBR file
*/
close_file()
{
    if (!library())
        return;

    printf(" Library %s closed\n", curr_file);
    fclose(LBR_fs);
    curr_file[0] = '\0';
}

/* extract_files()
extract library files to disk/directory
*/
extract_files(mask)
    char *mask;
{
    int last = 0, len, l;
    char dest[MAXPATH], outfn[MAXPATH+13];

    if (!mask[0] || mask == 1) return;

    if (!library()) return;

    printf("Extract to: ");
    input(dest, MAXPATH-2);

    if (dest[0] && dest[strlen(dest)-1] != 92)
        strcat(dest, "\\");

    strupr(dest);
    outfn[0] = '\0';

    while(last != ERROR) {
        last = find_member(mask, last);
        if (last != ERROR) {
            printf(" Extracting %.8s.%.3s   to %s ...", entry[last].name,
              entry[last].ext, dest);

            sprintf(outfn, "%s%.8s.%.3s", dest, entry[last].name,
              entry[last].ext);
            if ((aux_fs=fopen(outfn, "wb"))==NULL) {
                printf("\n CAN NOT OPEN OUTPUT FILE\n");
                return;
            }
            rewind(LBR_fs);
            fseek(LBR_fs, (long) (entry[last].index*128), SEEK_SET);

            len = entry[last].length;
            while(len) {
                fread(sector_buf, 128, 1, LBR_fs);

                l = 128;
                if (len == 1)
                    l -= entry[last].pad_count;

                if (fwrite(sector_buf, l, 1, aux_fs) != 1) {
                    printf(" \n ERROR WRITING! Disk full?\n");
                    fclose(aux_fs);
                    return;
                }
                len--;
            }
            fclose(aux_fs);
            putchar('\n');
        }
    }
    if (!outfn[0])
        puts(not_found);
}

/* crunch_file()
crunches library by expanding or shrinking to new size
*/
crunch_file()
{
    int p, last = 0, i, len;
    static char temp[20];

    if (!library()) return;

    p = index(curr_file, '.');
    strncpy(temp, curr_file, p);
    strcpy(&temp[p], ".$$$");

    if (rename(curr_file, temp) == ERROR) {
        printf(" CAN NOT CREATE TEMPORARY FILE\n");
        return;
    }

    printf("To crunch %s, please enter the\n", curr_file);
    printf("new number of members.\n");

    if (open_new(curr_file, aux_fs, active_members) == ERROR) {
        unlink(curr_file);
        rename(temp, curr_file);
        return;
    }

    i = 0;
    while(last != ERROR) {
        last = find_member("*.*", last);
        if (last != ERROR) {
            i++;
            printf(" Moving %.8s.%.3s to member %d", entry[last].name,
              entry[last].ext, i);
            rewind(LBR_fs);
            fseek(LBR_fs, (long) (entry[last].index*128), SEEK_SET);

            len = entry[last].length;

            if ((aux_fs=fopen(curr_file, "ab"))==NULL) {
                printf(" CAN NOT APPEND TO NEW LIBRARY!\n");
                unlink(curr_file);
                rename(temp, curr_file);
                return;
            }
            member_copy(last, 0);
            entry[0].index = (filelength(aux_fs->fd)/128);


            while(len) {
                fread(sector_buf, 128, 1, LBR_fs);

                if (fwrite(sector_buf, 128, 1, aux_fs) != 1) {
                    printf(" \n ERROR WRITING! Disk full?\n");
                    fclose(aux_fs);
                    unlink(curr_file);
                    rename(temp, curr_file);
                    return;
                }
                len--;
            }

            fclose(aux_fs);
            if ((aux_fs=fopen(curr_file, "r+b")) == NULL) {
                printf("\n ERROR OPENING NEW LIBRARY\n");
                unlink(curr_file);
                rename(temp, curr_file);
                return;
            }
            fseek(aux_fs, (long) (i*sizeof(entry[0])), SEEK_SET);
            if (fwrite(&entry[0], sizeof(entry[0]), 1, aux_fs) != 1) {
                printf(" ERROR WRITING INDEX\n");
                unlink(curr_file);
                rename(temp, curr_file);
                return;
            }
            fclose(aux_fs);
            putchar('\n');
        }
    }
    putchar('\n');

    fcloseall();
    unlink(temp);
    strcpy(temp, curr_file);
    curr_file[0] = '\0';
    open_file(temp);
}

/* rename_files()
renames members
*/
rename_files(old, new)
    char *old, *new;
{
    int i;
    char name[8], ext[3];

    if (!old[0] || !new[0] || old == 1 || new == 1) return;

    if (!library()) return;

    if (index(old, '?') != ERROR || index(old, '*') != ERROR ||
       index(new, '?') != ERROR || index(new, '*') != ERROR) {
         printf(" WILDCARDS NOT ALLOWED!\n");
         return;
    }

    strupr(old);
    strupr(new);

    i = find_member(old, 0);
    if (i == ERROR) {
        puts(not_found);
        return;
    }

    printf(" Renaming %s to %s", old, new);

    pad(name, ' ', 8);
    pad(ext, ' ', 3);

    fn_split(new, name, ext, FALSE);
    strncpy(entry[i].name, name, 8);
    strncpy(entry[i].ext, ext, 3);
    write_index(i);
    putchar('\n');

}

/* print_files()
prints files from library to stdprn
*/
print_files(mask)
    char *mask;
{
    int last = 0, len, l;

    if (!mask[0] || mask == 1) return;

    if (!library()) return;

    while(last != ERROR) {
        last = find_member(mask, last);
        if (last != ERROR) {
            printf(" Printing %.8s.%.3s", entry[last].name, entry[last].ext);

            rewind(LBR_fs);
            fseek(LBR_fs, (long) (entry[last].index*128), SEEK_SET);

            len = entry[last].length;
            while(len) {
                fread(sector_buf, 128, 1, LBR_fs);

                l = 128;
                if (len == 1)
                    l -= entry[last].pad_count;

                if (fwrite(sector_buf, l, 1, stdprn) != 1) {
                    printf(" \n ERROR WRITING! Printer on-line?\n");
                    return;
                }
                len--;
            }
            putchar('\n');
            fputc(12, stdprn);
        }
    }
    if (!last)
        puts(not_found);
}

/* replace_files()
replaces existing files with those outside the library,
if the file is the same size or smaller, the same space is used,
otherwise the file is appended
*/
replace_files(mask)
    char *mask;
{
    char done;
    int j, p, last, new;
    unsigned int l, d;
    struct ffblk dir;

    if (!mask[0] || mask == 1) return;

    if (!library()) return;

    done = findfirst(mask, &dir, 0);

    while(!done) {

        if ((last=find_member(dir.ff_name, 0)) != ERROR) {

            printf("  Replacing: %s", dir.ff_name);

            new = (dir.ff_fsize > entry[last].length * 128);
            l = dir.ff_fsize/128;
            if (dir.ff_fsize > l * 128) {
                l++;
                entry[last].pad_count = ((l) * 128) - dir.ff_fsize;
            }

            if ((aux_fs=fopen(dir.ff_name, "rb"))==NULL) {
                printf(" \n CAN NOT OPEN FILE!\n");
                return;
            }

            if (new) {
                fclose(LBR_fs);
                if ((LBR_fs=fopen(curr_file, "ab"))==NULL) {
                    printf(" CAN NOT APPEND TO LIBRARY!\n");
                    return;
                }
                entry[last].index = filelength(LBR_fs->fd)/128;
            }
            entry[last].length = l;


            if (!new) {
                rewind(LBR_fs);
                fseek(LBR_fs, (long) (entry[last].index*128), SEEK_SET);
            }

            while(TRUE) {
                strnset(sector_buf, 0x1A, 128);
                l = fread(sector_buf, 128, 1, aux_fs);

                if (fwrite(sector_buf, 128, 1, LBR_fs) != 1) {
                    printf(" \n ERROR WRITING! Disk full?\n");
                    fclose(aux_fs);
                    return;
                }
                if (l != 1) break;
            }
            fclose(aux_fs);
            putchar('\n');

            fclose(LBR_fs);
            if ((LBR_fs=fopen(curr_file, "r+b"))==NULL) {
                printf(" CAN NOT REOPEN LIBRARY!\n");
                close_file();
                return;
            }

            if (write_index(last) == ERROR) return;
        }
        done = findnext(&dir);

    }
}

/* rep_add_files()
replaces files specified, but if they don't exist, it will add_files()
them
*/
rep_add_files(mask)
    char *mask;
{
    char done;
    struct ffblk dir;

    if (!mask[0] || mask == 1) return;

    if (!library()) return;

    done = findfirst(mask, &dir, 0);

    while(!done) {

        if (find_member(dir.ff_name, 0) != ERROR) {
            replace_files(dir.ff_name);
        } else {
            add_files(dir.ff_name);
        }
        done = findnext(&dir);

    }
}

/* view_files()
views files by sending them to stdout
*/
view_files(mask)
    char *mask;
{
    int last = 0, len, l, key;

    if (!mask[0] || mask == 1) return;

    if (!library()) return;

    while(last != ERROR) {
        last = find_member(mask, last);
        if (last != ERROR) {
            printf(" Viewing file %.8s.%.3s   ^S=Pause, ^Q=Resume, ^X or ESC=Exit, ^N=Next\n",
                entry[last].name, entry[last].ext);
            printf("--------------------------------------------------------------------------\n");

            rewind(LBR_fs);
            fseek(LBR_fs, (long) (entry[last].index*128), SEEK_SET);

            len = entry[last].length;
            while(len) {
                if (kbhit()) {
                    key = getch();
                    if (key == 19) ungetch(key); /* ^S, save for later */
                    if (key == 14) break; /* ^N, next file */
                    if (key == 27 || key == 24) { /* ESC or ^X, can the puppy */
                        printf("\n\n");
                        return;
                    }
                }

                fread(sector_buf, 128, 1, LBR_fs);

                l = 128;
                if (len == 1)
                    l -= entry[last].pad_count;

                if (fwrite(sector_buf, l, 1, stdout) != 1) {
                    printf(" \n ERROR! STDOUT is not open!\n");
                    return;
                }
                len--;
            }
            printf("\n--------------------------------------------------------------------------\n");
        }
    }
    if (!last)
        puts(not_found);
}

/* delete_files()
removes files from library by tagging them as inactive
*/
delete_files(mask)
    char *mask;
{
    int last = 0;
    char match = FALSE;

    if (!mask[0] || mask == 1) return;

    while(last != ERROR) {
        last = find_member(mask, last);
        if (last != ERROR) {
            match = TRUE;
            printf(" Deleting %.8s.%.3s", entry[last].name,
              entry[last].ext);
            entry[last].status = DELETED;
            write_index(last);
            putchar('\n');
            active_members--;
        }
    }
    if (!match)
        puts(not_found);
}

/* open_file()
open library file if there isn't one opened already
*/
open_file(filename)
    char *filename;
{
    unsigned int filelen;
    int i, j, deleted, unused, fill;
    static char fn[14];

    strcpy(fn, filename);

    if (strpbrk(fn, "/\\:") != NULL) {
        printf("\n LIBRARY FILE MUST BE IN CURRENT PATH!\n");
        printf(" Use -U to change current logged path\n");
        return;
    }

    if (!fn[0] || filename == 1) return;

    if (curr_file[0]) { /* one is already open */
        printf(" LIBRARY IS ALREADY OPEN\n");
        sound(200);
        delay(100);
        nosound();
        return;
    }

    if (strchr(fn, '.')==NULL)
        strcat(fn, ".LBR");

    if ((LBR_fs=fopen(fn, "r+b"))==NULL) {
        printf("\n  LIBRARY DOES NOT EXIST\n");
        printf(" To create the library, enter the number of\n");
        printf("members, or <ENTER> to abort its creation.\n\n");
        if (open_new(fn, LBR_fs, 1) == ERROR) return;
        LBR_fs = fopen(fn, "r+b");
    }

    rewind(LBR_fs);
    if (fread(&entry[0], sizeof(entry[0]), 1, LBR_fs) != 1) {
        printf(" ERROR READING TOC\n");
        fclose(LBR_fs);
        return;
    }

    filelen = (entry[0].length * 4);
    if (filelen > 510) {
        printf(" LIBRARY IS TOO LARGE!\n 510 IS MAXIMUM FOR ALU\n");
        fclose(LBR_fs);
        return;
    }
    for (i=0; i < filelen; i++) {
        fseek(LBR_fs, (long) (i*sizeof(entry[0])), SEEK_SET);
        if (fread(&entry[i], sizeof(entry[0]), 1, LBR_fs) != TRUE) {
            printf(" ERROR READING INDEX\n");
            fclose(LBR_fs);
            return;
        }
    }

    members = i;
    for (j=i=deleted=unused=0; j<members; j++) {
        if (entry[j].status == ACTIVE)
            i++;
        else if (entry[j].status == UNUSED)
            unused++;
        else if (entry[j].status != ACTIVE)
            deleted++;
    }
    strcpy(curr_file, strupr(fn));
    printf(" Library %s open\n", curr_file);
    active_members = i;
    printf(" %d entries, %d active, %d deleted, %d unused\n\n", members,
      i, deleted, unused);
    rewind(LBR_fs);
}

/* get_path()
inputs new path (drive/dir) from user and logs to it,
will reset disk if we're logging to a new drive, and
look for currently logged dir if no dir is specified
*/
get_path(path)
    char *path;
{
    char temp_path[MAXPATH];
    int p;

    if (curr_file[0]) {
        printf(" CAN NOT CHANGE PATH WITH LIBRARY OPEN\n");
        return;
    }

    if (!path[0] || path == 1) {
        printf("New path: ");
        input(temp_path, MAXPATH-2);
        if (!temp_path[0]) return;
    } else {
        strcpy(temp_path, path);
    }

    strupr(temp_path);

    p = strlen(temp_path)-1;

    if (temp_path[p] == 92) /* backslash is no-no with */
        temp_path[p] = '\0'; /* chdir() */

    if (temp_path[p] != ':') {
        if (chdir(temp_path) == ERROR) {
            printf(" INVALID PATH\n");
            return;
        }
    }

    if (temp_path[1] == ':' && old_drive[0] != temp_path[0]) {
        setdisk(temp_path[0]-'A');
        old_drive[0] = temp_path[0];
    }

    getcwd(curr_path, MAXPATH);
}

/* do_dir()
display directory of currently logged drive/dir
*/
do_dir(path)
    char *path;
{
    char done, files;
    struct ffblk dir;

    if (!path[0] || path == 1)
        strcpy(path, "*.*");

    files = 0;

    done = findfirst(path, &dir, 0);

    while(!done && !kbhit()) {
        if (!files)
            putchar('\n');
        files++;
        printf("%-13s %8ld\n", dir.ff_name, dir.ff_fsize);
        done = findnext(&dir);
    }

    while(kbhit())
        getch();

    if (!files)
        printf(" NO FILES FOUND");
    putchar('\n');
}

/* write_index()
writes specified index to TOC
*/
int write_index(i)
    int i;
{

    fseek(LBR_fs, (long) (i*sizeof(entry[0])), SEEK_SET);
    if (fwrite(&entry[i], sizeof(entry[0]), 1, LBR_fs) != 1) {
        printf(" ERROR WRITING INDEX\n");
        return(ERROR);
    }
    return(FALSE);
}

/* open_new()
opens and creates new library
*/
open_new(fn, fs, min)
    char *fn;
    FILE *fs;
    int min;
{
    char str[5];
    int i, j, fill, t;

    printf("Enter number of members: ");
    input(str, 4);
    if (!str[0]) return(ERROR);
    i = atoi(str);
    if (i < min || i > 512) return(ERROR);
    i++; /* directory entry */
    j = i/4;
    if (j * 4 < i) {
        j++;
    }
    clear_member(&entry[0], ACTIVE);
    entry[0].length = j;
    if ((fs=fopen(fn, "wb"))==NULL) {
        printf(" CAN NOT OPEN LIBRARY!\n");
        return(ERROR);
    }
    if (fwrite(&entry[0], sizeof(entry[0]), 1, fs) != 1) {
        printf(" ERROR WRITING TOC\n");
        fclose(fs);
        return(ERROR);
    }
    t = sizeof(entry[0]) * (j*4);
    j = t/128;
    fill = 0;
    if (j * 128 < t) {
        fill = (j+1) * 128 - t;
    }
    clear_member(&entry[511], UNUSED);
    for (i=1; i<j*4; i++) {
        if (fwrite(&entry[511], sizeof(entry[0]), 1, fs) != 1) {
            printf(" ERROR WRITING INDEX\n");
            fclose(fs);
            return(ERROR);
        }
    }
    if (fill) {
        for (i=0; i<fill; i++)
            fputc(0x1A, fs);
    }
    fclose(fs);

    return(FALSE);
}

